// Olden - Disassembly listing oldener for IDA v5.1
// Copyright 2007 (c) Ilfak Guilfanov <ig@hexblog.com>
// Freeware. Please keep the copyright message intact.
//
// This plugin runs in the background and modifies the color of the items
// visited by the user. It darkens/lightens the current line each second by several
// gradations until the target color is reached. The default target color
// is green.
//
// The net effect of this plugin is that the lines visited by the user
// many times or for a long time will become perceptibly darker/lighter.
// The question of 'Have I been here before?' can be answered more easily.
// Might be especially useful when analyzing huge applications.

#include <windows.h>
#include <ida.hpp>
#include <idp.hpp>
#include <bytes.hpp>
#include <loader.hpp>
#include <kernwin.hpp>

#pragma comment(lib, "advapi32")

//--------------------------------------------------------------------------
static const bgcolor_t default_target_color = 0x80FF80; // green

// pointer to kernel function to mark disassembly list for a refresh
void (idaapi*idaview_marker)(ea_t ea);

// Netnode name to save the plugin information
static const char nodename[] = "$ darkener";
// Node to keep the plugin information
static netnode node;

// Format of plugin settings
struct darkener_info_t
{
  bgcolor_t target_color;
  uint32 active;                // is plugin active?
};

static darkener_info_t di = { default_target_color, true };
static bgcolor_t default_bgcolor;

static const int ngrades = 64;  // number of color grades to use
static const bgcolor_t WHITE_COLOR = bgcolor_t(0xFFFFFF);

#pragma pack(push, 4)
struct gui_palette_info_t
{
  char magic[6];
  ushort enabled;
  ushort palnum;
  ulong palettes[4][PALETTE_SIZE];
};
#pragma pack(pop)

//--------------------------------------------------------------------------
static bgcolor_t get_default_bgcolor(void)
{
#ifdef __NT__
  HKEY hkey;
  DWORD disp;
  bgcolor_t color = WHITE_COLOR;
  if ( RegCreateKeyExA(HKEY_CURRENT_USER, "Software\\Datarescue\\IDA", 0, "",
                       REG_OPTION_NON_VOLATILE, KEY_READ, NULL,
                       &hkey, &disp) == ERROR_SUCCESS )
  {
    if ( disp == REG_OPENED_EXISTING_KEY )
    {
//    char buf[1024];
      gui_palette_info_t pal;
      DWORD type;
      DWORD sz = sizeof(pal);
//      DWORD sz = sizeof(buf);
//      DWORD code = RegQueryValueExA(hkey, "Palette", NULL, &type, buf, &sz);
      DWORD code = RegQueryValueExA(hkey, "Palette", NULL, &type, (LPBYTE)&pal, &sz);
      if ( code == ERROR_SUCCESS || code == ERROR_MORE_DATA )
      {
        if ( memcmp(pal.magic, "IDACLR", 6) == 0 )
        {
          if ( pal.palnum < 4 )
            color = pal.palettes[pal.palnum][41];
        }
      }
    }
    RegCloseKey(hkey);
  }
  return color;
#else
  return WHITE_COLOR;
#endif
}

//--------------------------------------------------------------------------
// Try to modify the color grade of the specified address
static bool increment_grade(ea_t ea)
{
  int cnt = node.altval(ea);
  if ( cnt >= ngrades )
    return false;

  node.altset(ea, cnt+1);
  return true;
}

//--------------------------------------------------------------------------
static bgcolor_t age_color(bgcolor_t color, int grade)
{
  bgcolor_t res = 0;
  for ( int i=0; i < 3; i++ )
  {
    uchar orig = uchar(color >> (i*8));
    uchar target = uchar(di.target_color >> (i*8));
    int delta = ((target - orig) * grade) / ngrades;
    orig = uchar(orig + delta);
    res |= bgcolor_t(orig) << (i*8);
  }
  return res;
}

//--------------------------------------------------------------------------
// User interface event callback. Here we darken addresses.
static int idaapi ui_callback(void * /*user_data*/, int code, va_list /*va*/)
{
  switch ( code )
  {
    // this event occurs quite often, so we check the current address
    // and change colors here.
    case ui_setstate:

      // plugin active?
      if ( !di.active )
        break;

      // enough time passed?
      static time_t last = 0;
      time_t now = time(NULL);
      if ( last+1 >= now )
        break;
      last = now;

      // valid current address in the listing?
      ea_t ea = get_screen_ea();
      if ( ea == BADADDR )
        break;

      // already reached the target color?
      if ( !increment_grade(ea) )
        break;

      // tell ida to refresh the screen
      mark_idaview_for_refresh(ea);
      break;
  }
  return 0;
}

//--------------------------------------------------------------------------
// Procossor event callback. Here supply the kernel with new dynamic colors.
static int idaapi idp_callback(void * /*user_data*/, int code, va_list va)
{
  switch ( code )
  {
    case processor_t::get_bg_color:
                                // Get item background color
                                // in: ea_t ea, bgcolor_t *color
                                // Returns: 1-not implemented, 2-color set
                                // Plugins can hook this callback to color disassembly lines
                                // dynamically
      ea_t ea = va_arg(va, ea_t);
      bgcolor_t *p_color = va_arg(va, bgcolor_t *);
      int cnt = node.altval(ea);
      if ( cnt > 0 )
      {
        // get original color
        bgcolor_t color = age_color(default_bgcolor, cnt);
        *p_color = color;
        return 2;
      }
      break;
  }
  return 0;
}

//--------------------------------------------------------------------------
// Reset all item colors in the disassembly
static void idaapi reset_colors(TView *[], int)
{
  node.altdel();
  info("All disassembly item colors have been reset");
}

//--------------------------------------------------------------------------
void run(int /*arg*/)
{
  // refresh the default background color
  default_bgcolor = get_default_bgcolor();
  // configure the plugin
  if ( AskUsingForm_c(
    "Olden\n"
    "\n"
    "<Plugin is active:C>>\n"
    "\n"
    "<#Reset all disassembly item colors#~R~eset listing colors:B:0:::>\n"
    "\n"
    "<#Please specify the final color#~T~arget color  :K::8::>\n"
    "\n",
    &di.active,
    reset_colors,
    &di.target_color) )
  {
    // save our data at the index 0
    node.supset(0, &di, sizeof(di));

    // setup the hook
    if ( di.active )
    {
      hook_to_notification_point(HT_UI, ui_callback, NULL);
      hook_to_notification_point(HT_IDP, idp_callback, NULL);
    }
    else
    {
      unhook_from_notification_point(HT_UI, ui_callback, NULL);
      unhook_from_notification_point(HT_IDP, idp_callback, NULL);
    }
  }
}

//--------------------------------------------------------------------------
int init(void)
{
  // works for gui version
  if ( callui(ui_get_hwnd).vptr == NULL )
    return PLUGIN_SKIP;

  // if there are custom plugin settings, retrieve them
  node.create(nodename);
  node.supval(0, &di, sizeof(di));

  // hook to the events
  if ( di.active )
  {
    default_bgcolor = get_default_bgcolor();
    hook_to_notification_point(HT_UI, ui_callback, NULL);
    hook_to_notification_point(HT_IDP, idp_callback, NULL);
  }

  // setup a pointer to the marker function. we will need it
  // to planify refreshs. this method is better than explicitly refreshing
  // the listing because the interface might accumulate changes and perform
  // refresh onyl once
  setup_idaview_marker();

  // ask the kernel to keep us in the memory since we have installed a hook
  return PLUGIN_KEEP;
}

//--------------------------------------------------------------------------
void term(void)
{
  // uninstall the hook before unloading from the memory
  if ( di.active )
  {
    unhook_from_notification_point(HT_UI, ui_callback, NULL);
    unhook_from_notification_point(HT_IDP, idp_callback, NULL);
  }
}

//--------------------------------------------------------------------------
char wanted_name[] = "Olden";
char wanted_hotkey[] = "";


//--------------------------------------------------------------------------
//
//      PLUGIN DESCRIPTION BLOCK
//
//--------------------------------------------------------------------------
plugin_t PLUGIN =
{
  IDP_INTERFACE_VERSION,
  0,                    // plugin flags
  init,                 // initialize

  term,                 // terminate. this pointer may be NULL.

  run,                  // invoke plugin

  wanted_name,          // long comment about the plugin
                        // it could appear in the status line
                        // or as a hint

  wanted_name,          // multiline help about the plugin

  wanted_name,          // the preferred short name of the plugin
  wanted_hotkey         // the preferred hotkey to run the plugin
};
